import os
import time

# BUILD FLAGS

# to suppress warnings for suspicious lines of code
# add "[buildlocales.py No Warning]" somewhere on the line
ALLOW_HIDING_WARNINGS = True

# [buildlocales.py Copy]
ALLOW_COPY = True

# scans the beginning of Locale-enUS.lua for mismatched definitions
# deprecated since the COPY_COMMAND_START is easier to use/maintain
ALLOW_MANUAL_DEFINITIONS = False

# run a validation check on the other local files to check for unused or missing translations
ALLOW_VALIDATE_TRANSATIONS = False

# Complain about undefined keys at runtime (an AceLocale-3.0 setting)
# It's useful to enable complaining for validating the scan
# But disabling it is useful for supporting incomplete translations
# use string of Lua keyword instead of Python keyword
ALLOW_COMPLAIN = "true" 

# command code strings to drive special behavior in code files
COMMAND_NO_WARNINGS = "[buildlocales.py No Warning]" # anywhere on the line
COMMAND_COPY_START = "-- [buildlocales.py copy]" # has to match the entire line
COMMAND_COPY_STOP = "-- [buildlocales.py end of copy]" # has to match the entire line


# File Lists (controls order of processing)
CODE_FILES = [
    #"loader.lua", #not allowed to use L[] per TOC file order
    #"debug.lua", #not allowed to use L[] per TOC file order
    "SpeakinSpell.lua", #list first for dominant definitions/usage
    "substitutions.lua", # where these are used in the gui, i want this file to be considered the owner
    "slashcommands.lua", # same
    "data.lua",
    "utils.lua",
    "parser.lua",
    "gui/gui.lua", #list first in the GUI group, and use GUI files first
    "gui/generaloptions.lua",
    "gui/currentmessages.lua",
    "gui/createnew.lua",
    "templates.lua",
    "gui/import.lua",
    "gui/colors.lua",
    "gui/help.lua",
    "gui/minimapbutton.lua",
    "gui/networkgui.lua",
    "gui/randomsubs.lua",
    #"gui\trade.lua",#not yet implemented, nor used
    #"clickhere.lua", #just doesn't use L[] but is allowed
    "speech.lua",
    "network.lua",
    "rez.lua",
    "wowevents.lua",
    #"ads.lua", #gets everything it needs from DefaultSpeeches-xxXX.lua
    "oldversions.lua", #gets special attention, scanned last for validation
    ]

TRANSLATIONS = [
    "Locales/Locale-deDE.lua",
    #"Locales/Locale-koKR.lua",#TODO: how to open this as unicode?
    ]


# constants
INCLUDE = """
local L = LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell", false)
""".strip()
MANUAL_DEFINITION = "manual definitions"


# Globals
g_AllFoundKeys = dict()
g_ManualDefs = list()
g_ManualDefsUsed = list()


def ShowWarning(warning,line):
    global ALLOW_HIDING_WARNINGS
    global COMMAND_NO_WARNINGS
    if ALLOW_HIDING_WARNINGS and line.find(COMMAND_NO_WARNINGS) >= 0:
        return
    print("-- WARNING: " + warning)
    return


def ScanManualDefs():
    f = open("Locales/Locale-enUS.lua")
    contents = f.readlines()
    f.close()
    global g_ManualDefs
    global MANUAL_DEFINITION
    for line in contents:
        if line.strip() == "-- AUTO-GENERATED LIST":
            return
        if line.startswith("L["):
            end = line.find("\"]")
            key = line[2:end+1]
            g_AllFoundKeys[key] = MANUAL_DEFINITION
            g_ManualDefs.append(key)
    return


def ReportUnusedManualDefs():
    global g_ManualDefs
    global g_ManualDefsUsed
    print("\n")
    print("-"*60)
    print("-- UNUSED MANUAL DEFINITIONS")
    print("-"*60)
    ReportedSomething = False
    for key in g_ManualDefs:
        if not key in g_ManualDefsUsed:
            print("-- %s"%key)
            ReportedSomething = True
    if not ReportedSomething:
        print("-- None! Yay!")
    print("\n")
    return


def ScanCodeFile(filename):
    # declarations
    global g_AllFoundKeys
    global INCLUDE
    global MANUAL_DEFINITION
    global ALLOW_HIDING_WARNINGS
    global COMMAND_NO_WARNINGS
    FoundKeys = list()
    FoundInclude = False
    MultiLine = False
    LineNum = 0
    CopyMode = False

    # check parameters
    if not os.path.isfile(filename):
        return
    if not filename.endswith(".lua"):
        return

    # read the file
    f = open(filename)
    contents = f.readlines()
    f.close()

    # print a header
    print("")
    print("-"*42)
    print( "-"*10+" " + "%-20s"%filename+" "+"-"*10 )
    print("-"*42)
    
    # parse for matches
    for line in contents:
        LineNum = LineNum+1
        #check for the declaration of L[]
        if line.strip() == INCLUDE:
            FoundInclude = True
            continue

        #check for commands
        if ALLOW_COPY and line.strip() == COMMAND_COPY_START:
            CopyMode = True
            #fall through to scan the copied text for L[] keys

        if CopyMode:
            print(line.strip())
            if line.strip() == COMMAND_COPY_STOP:
                CopyMode = False
                continue # there won't be anything else to find
            #fall through to scan the copied text for L[] keys
            
        #ignore comments
        if line.strip().startswith("--"):
            continue
        
        #check for multi-line tooltips
        if line.strip().endswith("L["):
            MultiLine = True
            Tooltip = ""
            TooltipStartLine = LineNum
            continue
        
        if MultiLine:
            Tooltip = Tooltip + line
            if line.strip().endswith("]]"):
                MultiLine = False
                if not Tooltip in FoundKeys:
                    FoundKeys.append( Tooltip )
                    if Tooltip in g_AllFoundKeys.keys():
                        firstuse = g_AllFoundKeys[Tooltip]
                        if firstuse == MANUAL_DEFINITION:
                            g_ManualDefsUsed.append(Tooltip)
                        if not (ALLOW_HIDING_WARNINGS and Tooltip.find(COMMAND_NO_WARNINGS) >= 0):
                            print( "--[[ already defined under:%s\n%s\n--]]" %
                               (firstuse,Tooltip) )
                    else:
                        g_AllFoundKeys[Tooltip] = filename
                        if not CopyMode:
                            print( """
L[
%s] = 
%s""" % (Tooltip,Tooltip) )
            continue #we're in multiline mode, so anything else we look for below is part of the tooltip
        
        #check for the simple case L["something"]="something"
        start = line.find("L[\"")
        if start >= 0:
            end = line.find("\"]")
            while end > 0 and end < start:
                end = line.find("\"]",end+1)
            if end > start:
                key = line[start+2:end+1]
                if not key in FoundKeys:
                    FoundKeys.append( key )
                    if key in g_AllFoundKeys.keys():
                        firstuse = g_AllFoundKeys[key]
                        if firstuse == MANUAL_DEFINITION:
                            g_ManualDefsUsed.append(key)
                        if not (ALLOW_HIDING_WARNINGS and line.find(COMMAND_NO_WARNINGS) >= 0):
                            print( "--L[%s] already defined under:%s" %
                               (key,firstuse) )
                    else:
                        g_AllFoundKeys[key] = filename
                        if not CopyMode:
                            print( "L[%s] = %s" % (key,key) )
            else: # start with no end, report something irregular
                if not CopyMode:
                    ShowWarning("unclosed L[\" in line:%s"%line.strip(), line)
            continue

        # check for any other irregularities
        start = line.find("L[")
        if start >= 0:
            if not CopyMode:
                ShowWarning("irregular usage of L[] in line:%s"%line.strip(),line)
            continue

        # else it must be a regular line of code we can ignore
        continue

    # check for declaration/include of L[]
    if not FoundInclude:
        ShowWarning("L[] include/declaration not found","")
    elif len(FoundKeys) == 0:
        ShowWarning("L[] is included/declared, but not used","")
    # check for unclosed multi-line syntax
    if MultiLine:
        ShowWarning("unclosed multi-line starting on line:%d"%TooltipStartLine,TooltipStartLine)
    return



def ValidateTranslation(filename):
    # check parameters
    if not os.path.isfile(filename):
        return
    if not filename.endswith(".lua"):
        return

    # read the file
    f = open(filename)
    contents = f.readlines()
    f.close()

    # print a header
    print("")
    print("-"*55)
    print( "-"*10+" VALIDATE: " + "%-20s"%filename+" "+"-"*10 )
    print("-"*55)

    # scan for unused translations
    Warned = False
    LineNum = 0
    FoundKeys = list()
    MultiLine = False
    for line in contents:
        LineNum = LineNum+1
        
        #ignore comments
        if line.strip().startswith("--"):
            continue
        
        #check for multi-line tooltips
        if line.strip().endswith("L["):
            MultiLine = True
            Tooltip = ""
            TooltipStartLine = LineNum
            continue
        
        if MultiLine:
            Tooltip = Tooltip + line
            if line.strip().endswith("]]"):
                MultiLine = False
                if not Tooltip in FoundKeys:
                    FoundKeys.append( Tooltip )
            continue #we're in multiline mode, so anything else we look for below is part of the tooltip
        
        #check for the simple case L["something"]="something"
        start = line.find("L[\"")
        if start >= 0:
            end = line.find("\"]")
            #this shouldn't occur in a Locale file
            #while end > 0 and end < start:
            #    end = line.find("\"]",end+1)
            if end > start:
                key = line[start+2:end+1]
                if not key in FoundKeys:
                    FoundKeys.append( key )
            else: # start with no end, report something irregular
                ShowWarning("unclosed L[\" in line:%s"%line.strip(),line)
            continue

        # check for any other irregularities
        start = line.find("L[")
        if start >= 0:
            ShowWarning("irregular usage of L[] in line:%s"%line.strip(),line)
            continue

        # else this is not a line we're looking for
        continue

    # check that all defined translations are needed
    global g_AllFoundKeys
    for definedkey in FoundKeys:
        if not definedkey in g_AllFoundKeys.keys():
            if not Warned:
                ShowWarning("Unnecessary translations:","")
                print("--[[")
            print(definedkey)
            Warned = True
    if Warned:
        print("--]]")
    else:
        print("-- PASS: No unused translations")
            
    # check that all needed translations are defined
    #TODO: start using this check when i think it will pass
    Warned = False
    for neededkey in g_AllFoundKeys.keys():
        if not neededkey in FoundKeys:
            if not Warned:
                ShowWarning("Missing translations:","")
                print("--[[")
            print(neededkey)
            Warned = True
    if Warned:
        print("--]]")
    else:
        print("-- PASS: Everything needed is defined")

    return


def main():
    # print header
    print(
"""-------------------------------------------------------------------------------
-- English localization file for enUS and enGB.
-- Auto-Generated by buildlocales.py on <%s>
-------------------------------------------------------------------------------

local AceLocale = LibStub:GetLibrary("AceLocale-3.0")
local complain = %s
local L = AceLocale:NewLocale("SpeakinSpell", "enUS", true, not complain)
if not L then return end

SpeakinSpell:PrintLoading("Locales/Locale-enUS.lua")"""%(time.asctime(),ALLOW_COMPLAIN))
    
    # init
    global ALLOW_MANUAL_DEFINITIONS
    if ALLOW_MANUAL_DEFINITIONS:
        ScanManualDefs() # COMMAND_COPY_START is better if possible
    
    # scan the files (shows results as it goes)
    global CODE_FILES
    for filename in CODE_FILES:
        ScanCodeFile( filename)
        
    # show more results
    if ALLOW_MANUAL_DEFINITIONS:
        ReportUnusedManualDefs() # COMMAND_COPY_START is better if possible
    
    # validate other locale files for completeness
    global ALLOW_VALIDATE_TRANSATIONS
    if ALLOW_VALIDATE_TRANSATIONS:
        global TRANSLATIONS
        for filename in TRANSLATIONS:
            ValidateTranslation(filename)
            
    print("") #I like a little white space at the end
    return


if __name__=="__main__":
    main()


